/* ful_util.c
 *   These have not been altered much since the original 
 *  version written by Crowell et al. I have streamlined the code a bit and 
 *  moved all the basic FUL functions to this file + "hidden" some functions
 *  that don't need to be global./KL
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ful_util.h"

#define MAX(x,y) ((x) > (y) ? x : y)
#define MIN(x,y) ((x) < (y) ? x : y)
#define SQUARED(x) ((x)*(x))

float old_ful(float *data, int size);
double narrow_chi_squared(float *data, int size, double k,
			  double max_diff, double est_fluct );
double modified_chi_squared(double flat_value, float *data, int size);

/*
 * modified_chi_squared (eq. 5)
 *
 * returns Sum_i((flat_value -data[i])^2)/Sum_i(flat_value)
 * where the first sum is carried out for data[i]<flat_value
 */

double modified_chi_squared(double flat_value, float *data, int size) {
  int   i;
  double mc;                 

  mc = 0.0;                  
  for(i = 0; i<size; i++){   
    if(data[i] < flat_value) 
      mc += SQUARED(flat_value - data[i]);
  }
  mc /= size * flat_value;   
  return mc;
}

/*
 * old_ful  - WAS do_ful
 *
 * given yi and m return a value for k
 */

float old_ful(float *data, int size) { 
  int         i;
  float  lowest;            
  float  lo_ful;            
  float  hi_ful;            
  float   lo_mc;            
  float   hi_mc;            
  float  interp;            
  float x_range;            
  float  new_mc;            
  float max_allowed_err;    

 
  i = 0;
  while (data[i++] < 0 && i < size)
    ;
  /*we should probably check that there is  at least one nonzero element*/
  lowest = data[i];         
  for(; i<size; i++){  
    if(data[i] < lowest && data[i] > 0)    
      lowest = data[i];     
  }
  lo_ful = MAX(1e-10,lowest);                              
  hi_ful = lowest + sqrt(size * lowest / 2.0);  

  lo_ful = MAX(1e-10,lo_ful);//=> lo_ful is positive
  hi_ful = MAX(0.5, hi_ful);//=> hi_ful is >= 0.5

  // next 2 lines - hi_ful was fabs(hi_ful). Removed - We have already
  // established that hi_ful >= 0.5. Also - should this be lo_ful
  // instead of lowest here?
  while(((1.0 / size) * SQUARED(hi_ful - lowest) / hi_ful) < 0.5){
    hi_ful = 2.0 * (hi_ful + 1); 
  } 

  /*
    Binary search for the correct value of ful. What is acceptable accuracy? 
    Well,  if we have n data-points and statistics/bin is about = ful,
    then error bars on ful should be about sqrt(ful)/sqrt(n). We'll demand
    .1 times this error,  or a relative error of .01,  whichever is lower. 
    */

  lo_mc = modified_chi_squared(lo_ful, data, size); 
  hi_mc = modified_chi_squared(hi_ful, data, size); 
 
  /*
    Do linear interpolation to look for the best guess as to where the 
    solution is,  but if linear interpolation yields something within 10% of 
    the ends of the interval,  don't trust it, because we might get stuck. 
    */
  while(1){
    if(lo_mc > 0.5){
      interp = lo_ful;
      //      interp = lo_mc; //I think this was a bug in the original code/KL
      break;
    }
    x_range = hi_ful - lo_ful;
    interp = lo_ful + (0.5 - lo_mc) * x_range / (hi_mc - lo_mc);
    // are we sure that the the following never will
    // result in an infinite loop? /KL
    interp = MIN(hi_ful - (0.1 *x_range), interp);
    interp = MAX(lo_ful + (0.1 *x_range), interp);

    new_mc = modified_chi_squared(interp, data, size);

    max_allowed_err = 0.1 * sqrt(interp / size);

    max_allowed_err = MIN(0.01 * interp, max_allowed_err);
    if( x_range < max_allowed_err)
      break;
    if(new_mc > 0.5){
      hi_ful = interp;
      hi_mc = new_mc;
    } else {
      lo_ful = interp;
      lo_mc = new_mc;
    }
  }
  return interp;
}

/*
 * narrow_chi_squared
 *
 * returns Sum_i(min(|data[i]-k|, max_diff)^2)/Sum_i(est_fluct^2)
 */
double narrow_chi_squared(float *data, int size, double k,
			  double max_diff, double est_fluct ) {
  int    i;
  double u, d;


  u = 0;                    
  for(i = 0; i < size; i++){  
    d = fabs(data[i] - k);  
    if( d > max_diff)       
      d = max_diff;        
    u += SQUARED(d);
  }
  return(u / (size *SQUARED(est_fluct)));
  
}

/* 
 * ful - WAS do_new_ful
 *
 * Should perhaps be renamed to something more reasonable.  This
 * function calls old_ful in order to get a first estimate for the FUL
 * value. The "best" FUL value is then found by means of chi-squared
 * minimization.  Is this two-step procedure done to take care of 
 * sparse data?/KL
 */

float ful(float *data, int size, float fluct_fact) {
  int    i, nn;
  double est_fluct;             
  double est_mean;              
  double lo_s;                  
  double hi_s;                  
  double z;                     
  double lo_ncs, hi_ncs;        
  double interp_s, interp_ncs;  
  double y_max;                 
  float  oful;               
  float  nful;               
  double max_diff;


  oful   = old_ful(data, size);
  //  return oful;
  oful   = MAX(0.5, oful);          

  est_mean  = oful; 
  est_fluct = 2.0 * sqrt(oful); 

  lo_s = 0.0;
  z    = 0.0;

  for(i = 0; i < size; i++) {              
    if(fabs(data[i] - oful) > z)
      z = fabs(data[i] - oful);
  }

  hi_s = z / est_fluct + 1e3;

  max_diff = lo_s*est_fluct;
  lo_ncs=narrow_chi_squared(data, size, est_mean,max_diff, est_fluct);
  max_diff = hi_s*est_fluct;
  hi_ncs=narrow_chi_squared(data, size, est_mean,max_diff, est_fluct);

  // start value for interp_s - keeps the compiler happy
  interp_s = lo_s + (hi_s - lo_s) *(1.0 - lo_ncs) / (hi_ncs - lo_ncs);
  if(hi_ncs < 1.01){
    interp_s   = hi_s;
    interp_ncs = hi_ncs;
  }else{
    interp_s = lo_s + (hi_s - lo_s)*(1.0 - lo_ncs)/(hi_ncs - lo_ncs);
    max_diff = interp_s*est_fluct;
    interp_ncs=narrow_chi_squared(data, size, est_mean, max_diff, est_fluct);
    while((hi_ncs - lo_ncs) > 1.0){
      interp_s = lo_s + (hi_s - lo_s)*(1.0 - lo_ncs)/(hi_ncs - lo_ncs);
      max_diff = interp_s*est_fluct;
      interp_ncs=narrow_chi_squared(data, size, est_mean, max_diff, est_fluct);
      if(interp_ncs > 1.0){
	hi_s   = interp_s;
	hi_ncs = interp_ncs;
      } else {
	lo_s   = interp_s;
	lo_ncs = interp_ncs;
      }
      interp_s   = 0.5 * (lo_s + hi_s);
      max_diff = interp_s*est_fluct;
      interp_ncs=narrow_chi_squared(data, size, est_mean, max_diff, est_fluct);
      if(interp_ncs > 1.0){
	hi_s   = interp_s;
	hi_ncs = interp_ncs;
      } else {
	lo_s   = interp_s;
	lo_ncs = interp_ncs;
      }
    }
  }
  /* Now do the final averaging step */
  y_max = est_mean + interp_s * est_fluct;
  y_max = MAX(y_max, oful);

  nful = 0.0;
  nn = 0;
  
  for(i = 0; i < size; i++){
    if(data[i] <= y_max){
      nful += data[i];
      nn++;
    }
  }
  nful /= nn;

  return nful;

}


