/* ful_util.c
 *   These have not been altered much since the original 
 *  version written by Crowell et al. I have streamlined the code a bit and 
 *  moved all the basic FUL functions to this file + "hidden" some functions
 *  that don't need to be global./KL
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ful_util.h"

#define MAX(x,y) ((x) > (y) ? x : y)
#define MIN(x,y) ((x) < (y) ? x : y)
#define SQUARED(x) ((x)*(x))

float old_ful(float *data, int size);
double narrow_chi_squared(float *data, int size, double k,
			  double max_diff, double est_fluct);
double modified_chi_squared(double flat_value, float *data, int size);

/*
 * modified_chi_squared (eq. 5)
 *
 * returns Sum_i((flat_value -data[i])^2)/Sum_i(flat_value)
 * where the first sum is carried out for data[i]<flat_value
 */

/* rewritten   DCR 10 Nov 2005
 */

double modified_chi_squared(double flat_value, float *data, int size)
{
  int   i;
  double mc = 0.0;

  for (i = 0; i < size && data[i] < flat_value ; i++) {
    mc += SQUARED(flat_value - data[i]);
  }
  return (mc / ((float) size * flat_value));
}

/*
 * old_ful  - WAS do_ful
 *
 * given yi = data and m = size,  return a value for k = FUL
 */

float old_ful(float *data, int size)
{
  /* DCR: modified version; data[] MUST be SORTED low to high */

  float lowest, lo_ful, hi_ful, lo_mc, hi_mc, new_mc;
  float interp, x_range, max_allowed_err;

  lowest = data[0];
  if (lowest < 0.0) fprintf(stderr, "ERROR in old_ful: negative data values...");
  lo_ful = lowest = MAX(0.001,lowest);
  hi_ful = lowest + sqrt((size + 4) * (lowest + 4.0));
  /* DCR: adding 4 to lowest in the above line fixes things for low statistics,
     i.e. lowest ~ 0, but doesn't screw up for large values of lowest.
     Adding 4 to size likewise helps deal with small data samples.
  */

  /* Interpolate to search for the correct value of ful.
     If we have n data-points and statistics/bin is ~ ful,
     then error bars on ful should be about sqrt(ful)/sqrt(n). We'll demand
     0.1 times this error,  or a relative error of 0.01,  whichever is lower.
  */

  lo_mc = modified_chi_squared(lo_ful, data, size);
  hi_mc = modified_chi_squared(hi_ful, data, size);
  interp = lo_ful;
  while (lo_mc < 0.5) {
    /* If linear interpolation yields something within 10% of the ends
       of the interval, don't trust it, because we might get stuck.
    */
    x_range = hi_ful - lo_ful;
    interp = lo_ful + (0.5 - lo_mc) * x_range / (hi_mc - lo_mc);
    interp = MIN(hi_ful - (0.1 *x_range), interp);
    interp = MAX(lo_ful + (0.1 *x_range), interp);

    max_allowed_err = 0.1 * sqrt(interp / size);
    max_allowed_err = MIN(0.01 * interp, max_allowed_err);
    if (x_range < max_allowed_err) break;

    new_mc = modified_chi_squared(interp, data, size);
    if (new_mc > 0.5) {
      hi_ful = interp;
      hi_mc = new_mc;
    } else {
      lo_ful = interp;
      lo_mc = new_mc;
    }
  }
  return interp;
}

/*
 * narrow_chi_squared
 *
 * returns Sum_i(min(|data[i]-k|, max_diff)^2)/Sum_i(est_fluct^2)
 */
double narrow_chi_squared(float *data, int size, double k,
			  double max_diff, double est_fluct)
{
  int    i;
  double u = 0.0, d;

  for (i = 0; i < size; i++) {
    d = fabs(data[i] - k);
    if (d > max_diff) d = max_diff;
    u += SQUARED(d);
  }
  return (u / ((float) size * SQUARED(est_fluct)));
}

/* 
 * ful - WAS do_new_ful
 *
 * Should perhaps be renamed to something more reasonable.  This
 * function calls old_ful in order to get a first estimate for the FUL
 * value. The "best" FUL value is then found by means of chi-squared
 * minimization.  Is this two-step procedure done to take care of 
 * sparse data?/KL
 */

float ful(float *in_data, int size)
{
  int    i, j, nn = 0;
  double est_fluct, est_mean, lo_s, hi_s;
  double z, lo_ncs, hi_ncs, interp_s, interp_ncs, y_max;
  float  ful = 0.0;
  static float  data[1024];

  /* DCR: modified version; sort data[] from low to high */
  data[0] = in_data[0];
  for (i = 1; i < size; i++) {
    for (j = i; j > 0; j--) {
      if (data[j-1] < in_data[i]) break;
      data[j] = data[j-1];
    }
    data[j] = in_data[i];
  }

  // return old_ful(data, size);

  est_mean  = MAX(0.5, old_ful(data, size));
  est_fluct = sqrt(est_mean);
  if (est_fluct < 1.0) est_fluct = 1.0;

  lo_s = 1.01;            /* lower bound for max_diff/est_fluct */
  z = fabs(data[0] - est_mean);
  if (z < fabs(data[size-1] - est_mean)) z = fabs(data[size-1] - est_mean);
  hi_s = z / est_fluct + 0.1;  /* upper bound for max_diff/est_fluct */

  lo_ncs = narrow_chi_squared(data, size, est_mean, lo_s * est_fluct, est_fluct);
  hi_ncs = narrow_chi_squared(data, size, est_mean, hi_s * est_fluct, est_fluct);

  /* We want to interpolate max_diff = interp_s*est_fluct to make narrow_chi_squared = 1.0
     Then we will be able to use y_max = est_mean + max_diff to average the data
     and get a final value for FUL.
   */

  interp_s  = hi_s;
  while (hi_ncs > 1.01 && (hi_ncs - lo_ncs) > 0.5) {
    /* interp_s = lo_s + (hi_s - lo_s) * (1.0 - lo_ncs)/(hi_ncs - lo_ncs); */
    interp_s = 0.5 * (lo_s + hi_s);
    interp_ncs = narrow_chi_squared(data, size, est_mean, interp_s * est_fluct, est_fluct);
    if (interp_ncs > 1.0) {
      hi_s   = interp_s;
      hi_ncs = interp_ncs;
    } else {
      lo_s   = interp_s;
      lo_ncs = interp_ncs;
    }
  }

  /* Now do the final averaging step */
  y_max = est_mean + interp_s * est_fluct;
  for (i = 0; i < size && data[i] <= y_max; i++) {
    ful += data[i];
    nn++;
  }

  return (ful / (float) nn);
}


